'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.superstruct = exports.struct = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _cloneDeep = require('lodash/cloneDeep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _componentType = require('component-type');

var _componentType2 = _interopRequireDefault(_componentType);

var _types = require('./types');

var _types2 = _interopRequireDefault(_types);

var _errors = require('./errors');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Create a struct factory from a set of `options`.
 *
 * @param {Object} options
 * @return {Function}
 */

function superstruct() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var TYPES = _extends({}, _types2.default, options.types || {});

  /**
   * Define a scalar struct with a `schema` type string.
   *
   * @param {String} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function scalarStruct(schema, defaults) {
    var isOptional = schema.endsWith('?');
    var type = isOptional ? schema.slice(0, -1) : schema;
    var types = type.split(/\s*\|\s*/g);

    var fns = types.map(function (t) {
      var fn = TYPES[t];

      if (typeof fn !== 'function') {
        throw new Error('No struct validator function found for type "' + t + '".');
      }

      return fn;
    });

    return function (value) {
      if (!isOptional && value === undefined) {
        throw new _errors.ValueRequiredError({ type: type });
      }

      if (value !== undefined && !fns.some(function (fn) {
        return fn(value);
      })) {
        throw new _errors.ValueInvalidError({ type: type, value: value });
      }

      return value;
    };
  }

  /**
   * Define a list struct with a `schema` array.
   *
   * @param {Array} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function listStruct(schema, defaults) {
    if (schema.length !== 1) {
      throw new Error('List structs must be defined as an array with a single element, but you passed ' + schema.length + ' elements.');
    }

    schema = schema[0];
    var fn = struct(schema);
    var type = 'array';

    return function (value) {
      if (value === undefined) {
        throw new _errors.ValueRequiredError({ type: type });
      } else if ((0, _componentType2.default)(value) !== 'array') {
        throw new _errors.ValueInvalidError({ type: type, value: value });
      }

      var ret = value.map(function (v, index) {
        try {
          return fn(v);
        } catch (e) {
          var path = [index].concat(e.path);

          switch (e.code) {
            case 'value_invalid':
              throw new _errors.ElementInvalidError(_extends({}, e, { index: index, path: path }));
            default:
              if ('path' in e) e.path = path;
              throw e;
          }
        }
      });

      return ret;
    };
  }

  /**
   * Define an object struct with a `schema` dictionary.
   *
   * @param {Object} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function objectStruct(schema, defaults) {
    var structs = {};
    var type = 'object';

    for (var key in schema) {
      var fn = struct(schema[key]);
      structs[key] = fn;
    }

    return function (value) {
      var isUndefined = false;

      if (value === undefined) {
        isUndefined = true;
        value = {};
      } else if ((0, _componentType2.default)(value) !== 'object') {
        throw new _errors.ValueInvalidError({ type: type, value: value });
      }

      var ret = {};

      for (var _key in structs) {
        var s = structs[_key];
        var v = value[_key];
        var r = void 0;

        try {
          r = s(v);
        } catch (e) {
          var path = [_key].concat(e.path);

          switch (e.code) {
            case 'value_invalid':
              throw new _errors.PropertyInvalidError(_extends({}, e, { key: _key, path: path }));
            case 'value_required':
              throw isUndefined ? new _errors.ValueRequiredError({ type: type }) : new _errors.PropertyRequiredError(_extends({}, e, { key: _key, path: path }));
            default:
              if ('path' in e) e.path = path;
              throw e;
          }
        }

        if (_key in value) {
          ret[_key] = r;
        }
      }

      for (var _key2 in value) {
        if (!(_key2 in structs)) {
          throw new _errors.PropertyUnknownError({ key: _key2, path: [_key2] });
        }
      }

      return isUndefined ? undefined : ret;
    };
  }

  /**
   * Define a struct with `schema`.
   *
   * @param {Function|String|Array|Object} schema
   * @param {Any} defaults
   * @return {Function}
   */

  function struct(schema, defaults) {
    var s = void 0;

    if ((0, _componentType2.default)(schema) === 'function') {
      s = schema;
    } else if ((0, _componentType2.default)(schema) === 'string') {
      s = scalarStruct(schema, defaults);
    } else if ((0, _componentType2.default)(schema) === 'array') {
      s = listStruct(schema, defaults);
    } else if ((0, _componentType2.default)(schema) === 'object') {
      s = objectStruct(schema, defaults);
    } else {
      throw new Error('A struct schema definition must be a string, array or object, but you passed: ' + schema);
    }

    return function (value) {
      if (value === undefined) {
        value = typeof defaults === 'function' ? defaults() : (0, _cloneDeep2.default)(defaults);
      }

      return s(value);
    };
  }

  /**
   * Return the struct factory.
   */

  return struct;
}

/**
 * Export the factory and the factory creator.
 *
 * @type {Function}
 */

var struct = superstruct();

exports.default = struct;
exports.struct = struct;
exports.superstruct = superstruct;